
/**
 * Copyright (C) 2005-2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * debug section: 15
 */

#include "squidwall.h"

/**
 * routines for antivirus checking with clamav
 */

static int clamd_type=0;
#define CLAMD_NONE   0
#define CLAMD_LOCAL  1
#define CLAMD_TCP    2

static char *clamav_headline="clamav";
static int check_clamav_connect_socket(request_t *r);
static int check_clamav_connect_tcp(request_t *r);
static int check_clamav_create_stream(request_t *r);
static void check_clamav_disconnect(request_t *r);
static void check_clamav_getversion(request_t *r);

/**
 * check_clamav_connect_tcp - create connection to clamd via tcp
 *
 * @return  0 on success, -1 on failure
 */
static int check_clamav_connect_tcp(request_t *r)
{
	debug(15, 3, "check_clamav_connect_tcp()");

	r->fd_clamd = socket_tcp4();
	if (r->fd_clamd == -1) {
		return -1;
	}

	socket_connect4(r->fd_clamd, (char*)&clamd_ip, clamd_port);
	if (!socket_connected(r->fd_clamd)) {
		close(r->fd_clamd);
		return -1;
	}

	if (!io_fd(r->fd_clamd)) {
		close(r->fd_clamd);
		return -1;
	}
	
	return 0;
}


/**
 * check_clamav_connect_socket - create connection to a local clamd socket
 *
 * @return 0 on success, -1 on failure
 */
static int check_clamav_connect_socket(request_t *r)
{
	struct sockaddr_un un;

	debug(15,3, "check_clamav_connect_socket()");

	r->fd_clamd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (r->fd_clamd == -1) {
		return -1;
	}

	un.sun_family=AF_UNIX;
	un.sun_path[fmt_strn(un.sun_path, clamav_socket, sizeof(un.sun_path))]=0;
	if (connect(r->fd_clamd, (struct sockaddr*)&un, sizeof(un)) == -1) {
		close(r->fd_clamd);
		return -1;
	}

	return 0;
}


/**
 * check_clamav_create_stream - create stream connection
 *
 * @return 0 on success, -1 on failure
 */
static int check_clamav_create_stream(request_t *r)
{
	stralloc clamdtext;

	debug(15,3, "check_clamav_create_stream()");

	stralloc_init(&clamdtext);

	/* initialize the command connection */
	buffer_init(&r->clamd_in,   (int(*)())read, r->fd_clamd,
		r->clamd_iBUF, sizeof r->clamd_iBUF);
	buffer_init(&r->clamd_out, (int(*)())write, r->fd_clamd,
		r->clamd_oBUF, sizeof r->clamd_oBUF);

	io_wantwrite(r->fd_clamd); io_wait(); io_dontwantwrite(r->fd_clamd);
	buffer_putsflush(&r->clamd_out, "STREAM\n");

	io_wantread(r->fd_clamd); io_wait(); io_dontwantread(r->fd_clamd);
	if (buffer_getnewline_sa(&r->clamd_in, &clamdtext) == 0) {
		debug(15,1, "can't read response of clamd");
		stralloc_free(&clamdtext);
		return -1;
	}
	clamdtext.s[clamdtext.len]=0;

	if (!str_start(clamdtext.s, "PORT ")) {
		debug(15,1, "bad response from clamd, expected: PORT xxx");
		stralloc_free(&clamdtext);
		return -1;
	}

	if (scan_uint(clamdtext.s+5, &clams_port) == 0) {
		debug(15,1, "can't read port of clamd stream");
		stralloc_free(&clamdtext);
		return -1;
	}
	stralloc_free(&clamdtext);

	r->fd_stream=socket_tcp4();
	if (r->fd_stream == -1) {
		debug(15,1, "socket() for clamd failed");
		return -1;
	}

	/**
	 * open connection to clamd stream socket
	 * - clamd_ip defaults to 127.0.0.1, so it is also valid for socket
	 */
	socket_connect4(r->fd_stream, (char*)&clamd_ip, clams_port);
	if (!socket_connected(r->fd_stream)) {
		debug(15,1,"connect() to clamd failed");
		return -1;
	}

	/* initialize the stream connection */
	buffer_init(&r->clams_out, (int(*)())write, r->fd_stream,
		r->clams_oBUF, sizeof r->clams_oBUF);

	if (!io_fd(r->fd_stream)) {
		debug(15,1,"io_fd() failed");
		return -1;
	}

	return 0;
}


/**
 * check_clamav_disconnect - disconnect from clamav
 *
 * @return nothing
 */
static void check_clamav_disconnect(request_t *r)
{
	debug(15,3, "check_clamav_disconnect()");
	r->clamd_in.fd=-1;  buffer_close(&r->clamd_in);
	r->clamd_out.fd=-1; buffer_close(&r->clamd_out);
	io_close(r->fd_clamd);

	return;
}


/**
 * check_clamav_connect - create connection to clamav
 *
 * @return 0 on success, -1 on failure
 */
int check_clamav_connect(request_t *r)
{
	int j = -1;

	debug(15,3, "check_clamav_connect()");
	switch (clamd_type) {
	case CLAMD_NONE:
		j = check_clamav_connect_tcp(r);
		if (j == 0) {
			debug(15,3, "%s: via tcp", clamav_headline);
			memcache_reinit(r);
			clamd_type=CLAMD_TCP;
			break;
		}
		j = check_clamav_connect_socket(r);
		if (j == 0) {
			debug(15,3, "%s: via socket file", clamav_headline);
			memcache_reinit(r);
			clamd_type=CLAMD_LOCAL;
			break;
		}
		break;
	case CLAMD_LOCAL:
		j = check_clamav_connect_socket(r);
		if (j == -1) memcache_reinit(r);
		break;
	case CLAMD_TCP:
		j = check_clamav_connect_tcp(r);
		if (j == -1) memcache_reinit(r);
		break;
	}

	if (j == -1) {
		debug(15,3, "%s: not available", clamav_headline);
		clamd_type=CLAMD_NONE;
		return j;
	}

	/* create stream, and return result */
	j = check_clamav_create_stream(r);

	return j;
}


/**
 * check_clamav_put2stream - send one character to clamav stream
 *
 * @param ch  next char, which should be send to clamav
 * @return    nothing
 */
void check_clamav_put2stream(request_t *r, char ch)
{
	if (clamd_type == CLAMD_NONE)
		return;

	buffer_put(&r->clams_out, &ch, 1);
	return;
}


/**
 * check_clamav_checkresult - check the stream ...
 * - closes the data stream to clamav
 * - check the result
 *
 * @return 0 => no virus, 1 => virus found
 */
int check_clamav_checkresult(request_t *r)
{
	stralloc clamdtext;

	if (clamd_type == CLAMD_NONE)
		return 0; /* no scanning -> no virus ;) */

	debug(15,3, "check_clamav_check_result()");
	stralloc_init(&clamdtext);

	/* flush and close stream connection */
	buffer_flush(&r->clams_out);
	r->clams_out.fd=-1; buffer_close(&r->clams_out);
	io_close(r->fd_stream);

	/* read the result on the command connection */
	io_wantread(r->fd_clamd); io_wait(); io_dontwantread(r->fd_clamd);
	if (!stralloc_copys(&clamdtext, "")) die_nomem();
	while (buffer_getline_sa(&r->clamd_in, &clamdtext) != 0);
	clamdtext.s[clamdtext.len]=0;

	check_clamav_disconnect(r);

	/* "stream: OK\n" => len==11 => no virus! */
	if (clamdtext.len == 11) {
		stralloc_free(&clamdtext);
		return 0;
	}

	/* "stream: ....\n" - but only 'FOUND\n' means there is a VIRUS! */
	debug(15,3, "clamd(%s)", clamdtext.s+clamdtext.len-6);

	if (str_start(clamdtext.s+clamdtext.len-6, "FOUND\n")) {

		/* "stream: Worm.Sober.I FOUND\n" -> "Worm.Sober.I" */
		clamdtext.s[clamdtext.len-7]=0; /* remove " FOUND\n" */
		if (!stralloc_copys(&r->virus, clamdtext.s+8)) die_nomem();
		stralloc_free(&clamdtext);

		/* we have a virus, get current version of clamav */
		check_clamav_getversion(r);
	
		return 1;
	}

	/* some error, or a new unknown response */
	stralloc_free(&clamdtext);
	return 0;
}


/**
 * check_clamav_getversion - get current version of clamav
 *
 * @return 0 => no virus, 1 => virus found
 */
void check_clamav_getversion(request_t *r)
{
	int j = -1;

	debug(15,3, "check_clamav_getversion()");
	switch (clamd_type) {
	case CLAMD_LOCAL:
		j = check_clamav_connect_socket(r);
		break;
	case CLAMD_TCP:
		j = check_clamav_connect_tcp(r);
		break;
	}

	/* some problem while connecting to clamav ?! */
	if (j == -1) {
		clamd_type=CLAMD_NONE;
		return;
	}

	/* get curent version of clamav */
	buffer_init(&r->clamd_in, (int(*)())read, r->fd_clamd,
		r->clamd_iBUF, sizeof r->clamd_iBUF);
	buffer_init(&r->clamd_out, (int(*)())write, r->fd_clamd,
		r->clamd_oBUF, sizeof r->clamd_oBUF);

	io_wantwrite(r->fd_clamd); io_wait(); io_dontwantwrite(r->fd_clamd);
	buffer_putsflush(&r->clamd_out, "VERSION\n");

	io_wantread(r->fd_clamd); io_wait(); io_dontwantread(r->fd_clamd);
	if (buffer_getnewline_sa(&r->clamd_in, &r->clamav) == 0) {
		close(r->fd_clamd);
		panic("can't read response of clamd");
		return;
	}
	r->clamav.len--; /* removes the \n */
	r->clamav.s[r->clamav.len]=0;

	if (!str_start(r->clamav.s, "ClamAV ")) {
		close(r->fd_clamd);
		panic("bad response from clamd, expected: 'ClamAV ' ...");
		return;
	}
	check_clamav_disconnect(r);

	return;
}
