
/**
 * Copyright (C) 2005-2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * debug section: 8 (logging / debugging)
 */

/**
 * logging / debugging functions
 */

#include "squidwall.h"
#include "timestamp.h"
#include "fmtX.h"

/* mutex for pid and fd_log */
static pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER;

static pid_t pid;
static int64 fd_log;

/**
 * log_init - initialize logfile and it's buffering
 *
 * @param r       the current thread request structure
 * @param reinit  should the logfile re-opened (1=yes)
 */
void log_init(request_t *r, int reinit)
{
	if (!logfile) return; /* logging disabled!!! */

	pthread_mutex_lock(&log_mutex);
	if (reinit)
	if (!io_appendfile(&fd_log, logfile)) {
		pthread_mutex_unlock(&log_mutex);
		logfile=0;
		return;
	}
	memset(r->log_BUF, 0, sizeof r->log_BUF);
	buffer_init(&r->logbuf, log_write, fd_log, r->log_BUF, sizeof r->log_BUF);
	pid=getpid();
	pthread_mutex_unlock(&log_mutex);

	return;
}

void log_timestamp(request_t *r)
{
	char s[TIMESTAMP];

	timestamp(s);
	buffer_put(&r->logbuf, s, TIMESTAMP);
	buffer_puts(&r->logbuf, " ");

	return;
}

void log_s(request_t *r, char *s)
{
	buffer_puts(&r->logbuf, s);

	return;
}

void log_sa(request_t *r, stralloc *sa)
{
	buffer_putsa(&r->logbuf, sa);

	return;
}

void log_flush(request_t *r)
{
	buffer_flush(&r->logbuf);

	return;
}

void log_u32pw(request_t *r, const unsigned long u, const unsigned int pad, const char with)
{
	char s[pad+1];

	s[fmt_u32pw(s, u, pad, with)]=0;
	buffer_puts(&r->logbuf, s);

	return;
}

void log_pid(request_t *r)
{
	log_u32pw(r, pid, 5, ' ');

	return;
}

void log_wid(request_t *r)
{
	log_u32pw(r, r->wid, 2, '0');

	return;
}

void log_headline(request_t *r, char *headline)
{
	log_timestamp(r);
	log_s(r, headline);

	log_s(r, "["); log_pid(r); log_s(r, "]");
	
	/* only usefull for the real threadpool */
	#if 0
	if (r->wid) {
		log_s(r, "["); log_wid(r); log_s(r, "]");
	}
	#endif

	return;
}

int log_write(int fd, char *s, unsigned int len)
{
	int r;

	if (!logfile) return 0; /* happens on bad logfile permissions */

	pthread_mutex_lock(&log_mutex);
	r=write(fd, s, len); /* XXX ... what is with error checking ?! */
	pthread_mutex_unlock(&log_mutex);

	return r;
}
