
/**
 * Copyright (C) 2005-2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * debug section: 3 (reading acl's)
 */

#include "squidwall.h"
#include "vline.h"

/* these 3 arrays are global */
array a_cfilter_todo, a_pfilter_todo;

static void read_acls_regex_options(struct named_regex *x,
		const char *path, char *file);
static void read_acls_regex(array *root, const char *path, char *file);
static void read_acls_string(array *root, const char *path, char *file);
static void read_acls_redirects(array *root, const char *path, char *file);
static void read_acls_phrase(array *root, const char *path, char *file);

static void read_acls_globals_default(void);

static void acl_init(void);
static void acl_reset(void);
static void free_array(array *root, int type);
static int read_dir2array(char *info, char *path, array *root, time_t *mtime,
		struct list_head *list, int type);

/* lists */
LIST_HEAD(hosts);
LIST_HEAD(users);
LIST_HEAD(globals);
LIST_HEAD(redirects);

static array a_hosts, a_users, a_globals, a_redirects;
static time_t t_hosts, t_users, t_globals, t_redirects;

LIST_HEAD(bfilter);
LIST_HEAD(brfilter);
LIST_HEAD(bfilter_url);
LIST_HEAD(bfilter_dom);
LIST_HEAD(bfilter_md5);
LIST_HEAD(bfilter_sha1);

LIST_HEAD(ufilter);
LIST_HEAD(urfilter);
LIST_HEAD(ufilter_url);
LIST_HEAD(ufilter_dom);
LIST_HEAD(ufilter_md5);
LIST_HEAD(ufilter_sha1);

LIST_HEAD(wfilter);
LIST_HEAD(wrfilter);
LIST_HEAD(wfilter_url);
LIST_HEAD(wfilter_dom);
LIST_HEAD(wfilter_md5);
LIST_HEAD(wfilter_sha1);

LIST_HEAD(pfilter);
LIST_HEAD(cfilter);
LIST_HEAD(crfilter);
LIST_HEAD(bcfilter);

static array
	a_bfilter, a_brfilter, a_bfilter_url,
	a_bfilter_dom, a_bfilter_md5, a_bfilter_sha1,
	a_ufilter, a_urfilter, a_ufilter_url,
	a_ufilter_dom, a_ufilter_md5, a_ufilter_sha1,
	a_wfilter, a_wrfilter, a_wfilter_url,
	a_wfilter_dom, a_wfilter_md5, a_wfilter_sha1,
	a_pfilter, a_cfilter, a_crfilter, a_bcfilter;

static time_t
	t_bfilter, t_brfilter, t_bfilter_url,
	t_bfilter_dom, t_bfilter_md5, t_bfilter_sha1,
	t_ufilter, t_urfilter, t_ufilter_url,
	t_ufilter_dom, t_ufilter_md5, t_ufilter_sha1,
	t_wfilter, t_wrfilter, t_wfilter_url,
	t_wfilter_dom, t_wfilter_md5, t_wfilter_sha1,
	t_pfilter, t_cfilter, t_crfilter, t_bcfilter;

struct squidwall_acl *acl;
struct squidwall_phrase *phrase;
struct squidwall_string *string;
struct squidwall_regex *regex;
struct named_regex *nr;

static void read_acls_regex_options(struct named_regex *x,
		const char *path, char *file)
{
	stralloc sa;
	buffer b;
	int plus=0;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;
		sa.s[sa.len] = 0; /* ensure termination */

		/* remove \r\n and \n */
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;

		switch (sa.s[0]) {
		case '+':
			plus = 1;
			break;
		case '-':
			plus = 0;
			break;
		default:
			continue;
		}

		if (str_start(sa.s+1, "REG_EXTENDED") && plus) {
			set(x->cflags, REG_EXTENDED);
		} else {
			unset(x->cflags, REG_EXTENDED);
		}

		if (str_start(sa.s+1, "REG_ICASE") && plus) {
			set(x->cflags, REG_EXTENDED);
		} else {
			unset(x->cflags, REG_EXTENDED);
		}

		if (str_start(sa.s+1, "REG_NOSUB") && plus) {
			set(x->cflags, REG_NOSUB);
		} else {
			unset(x->cflags, REG_NOSUB);
		}

		if (str_start(sa.s+1, "REG_NEWLINE") && plus) {
			set(x->cflags, REG_NEWLINE);
		} else {
			unset(x->cflags, REG_NEWLINE);
		}

		/* eflags for regexec() */
		if (str_start(sa.s+1, "REG_NOTBOL") && plus) {
			set(x->eflags, REG_NOTBOL);
		} else {
			unset(x->eflags, REG_NOTBOL);
		}

		if (str_start(sa.s+1, "REG_NOTEOL") && plus) {
			set(x->eflags, REG_NOTEOL);
		} else {
			unset(x->eflags, REG_NOTEOL);
		}

	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_acls_regex(array *root, const char *path, char *file)
{
	stralloc sa;
	buffer b;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	regex=array_allocate(root, sizeof(struct squidwall_regex),
			array_length(root, sizeof(struct squidwall_regex)));

	if (!regex) die_nomem();
	INIT_LIST_HEAD(&regex->list);

	/* save the filename of the filter */
	stralloc_init(&regex->filename);
	if (!stralloc_copys(&regex->filename, file)) die_nomem();
	regex->filename.s[regex->filename.len]=0;

	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;
		sa.s[sa.len] = 0; /* ensure termination */

		/* remove \r\n and \n */
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;

		/* save the precompiled regex */
		nr=array_allocate(&regex->regex, sizeof(struct named_regex),
				array_length(&regex->regex, sizeof(struct named_regex)));

		if (!nr) die_nomem();
		read_acls_regex_options(nr, path, "@");
		if (regcomp(&nr->r, sa.s, nr->cflags) != 0) {
			debug(3,3, "error: regcomp() failed @ this line: [%s]\n", sa.s);
			continue; /* read next line */
		}

		/* save the string */
		stralloc_init(&nr->n);
		if (!stralloc_copys(&nr->n, sa.s)) die_nomem();
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_acls_string(array *root, const char *path, char *file)
{
	stralloc sa;
	buffer b;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	/* add string to filter */
	string=array_allocate(root, sizeof(struct squidwall_string),
			array_length(root, sizeof(struct squidwall_string)));

	if (!string) die_nomem();
	INIT_LIST_HEAD(&string->list);

	/* save the filename of the filter */
	stralloc_init(&string->filename);
	if (!stralloc_copys(&string->filename, file)) die_nomem();
	string->filename.s[string->filename.len]=0;

	/* save all the strings to search for in the stralloc 'strings' */
	stralloc_init(&string->strings);
	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;
		sa.s[sa.len] = 0; /* ensure termination */

		/* remove \r\n and \n */
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;

		if (!stralloc_cats(&string->strings, sa.s)) die_nomem();
		if (!stralloc_0(&string->strings)) die_nomem();
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_acls_redirects(array *root, const char *path, char *file)
{
	stralloc fn, lnk;

	stralloc_init(&fn);
	stralloc_init(&lnk);

	if (!stralloc_copys(&fn, path)) die_nomem();
	if (!stralloc_cats(&fn, "/")) die_nomem();
	if (!stralloc_cats(&fn, file)) die_nomem();
	fn.s[fn.len]=0;

	/* space for readlink(2) */
	if (!stralloc_ready(&lnk, BUFSIZE)) die_nomem();
	if (readlink(fn.s, lnk.s, BUFSIZE-1) == -1) {
		/* we have nothing allocated ... just ignore this */
		stralloc_free(&fn);
		stralloc_free(&lnk);
		return;
	}

	/* add string to filter */
	string=array_allocate(root, sizeof(struct squidwall_string),
			array_length(root, sizeof(struct squidwall_string)));

	if (!string) die_nomem();
	INIT_LIST_HEAD(&string->list);

	/* save the filename (user url) of the filter */
	stralloc_init(&string->filename);
	if (!stralloc_copys(&string->filename, file)) die_nomem();
	string->filename.s[string->filename.len]=0;

	/* save now the wanted redirection site */
	stralloc_init(&string->strings);
	if (!stralloc_copys(&string->strings, lnk.s)) die_nomem();
	string->strings.s[string->strings.len]=0;

	stralloc_free(&fn);
	stralloc_free(&lnk);

	return;
}

static void read_acls_phrase(array *root, const char *path, char *file)
{
	stralloc sa;
	buffer b;
	int state, count=0;
	char *str, *num;
	unsigned int i, *ui;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	/* add string to filter */
	phrase=array_allocate(root, sizeof(struct squidwall_phrase),
			array_length(root, sizeof(struct squidwall_phrase)));

	if (!phrase) die_nomem();
	INIT_LIST_HEAD(&phrase->list);

	/* save the filename of the filter */
	stralloc_init(&phrase->filename);
	if (!stralloc_copys(&phrase->filename, file)) die_nomem();
	phrase->filename.s[phrase->filename.len]=0;

	/* save all the strings to search for in the stralloc 'strings' */
	stralloc_init(&phrase->strings);
	if (array_bytes(&phrase->points))
		array_reset(&phrase->points);
	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;
		sa.s[sa.len] = 0; /* ensure termination */

		/* remove \r\n and \n */
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;

		/* parsing of "< string ><num>" */
		for (i=0, state=0, num=0, str=0; i<=sa.len; i++) {
			switch (state) {
			case 0:
			if (sa.s[i] == '<') {
				sa.s[i]=0; state=1; str=sa.s+i+1;
			}
			break;
			case 1:
			if (sa.s[i] == '>') {
				sa.s[i]=0; state=2; sa.s[i]=0;
			}
			break;
			case 2:
			if (sa.s[i] == '<') {
				sa.s[i]=0; state=3; num=sa.s+i+1;
			}
			break;
			case 3:
			if (sa.s[i] == '>') {
				sa.s[i]=0; state=4; sa.s[i]=0;
			}
			break;
			} /* switch() */
		}

		if (state == 4) {
			/* sth. like "<lala><33>" --> was ok */
			ui=array_allocate(&phrase->points, sizeof(unsigned int), count);
			if (!ui) die_nomem();
			if (!scan_uint(num, ui)) *ui = PFILTER_DEFAULT;
			if (!stralloc_cats(&phrase->strings, str)) die_nomem();
			if (!stralloc_0(&phrase->strings)) die_nomem();
			count++;
		}
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_acls_globals_default(void)
{
	if (array_bytes(&a_globals)) return;

	/* if there no global set, allocate some dummy */
	acl=array_allocate(&a_globals, sizeof(struct squidwall_acl), 0);
	if (!acl) die_nomem();

	/* we allow all, if no globals are configured */
	acl->flags=ACL_ALLOW;
	acl->todo=TODO_DEFAULT;
	
	acl_init();
	acl_reset();

	return;
}

static void free_array(array *root, int type)
{
	int i,j;

	switch (type) {
	case TYPE_GLOBAL:
	case TYPE_ACL:
		for (i=0; i < array_length(root, sizeof(struct squidwall_acl)); i++) {
			acl = array_get(root, sizeof(struct squidwall_acl), i);
			stralloc_free(&acl->bfilter);
			stralloc_free(&acl->brfilter);
			stralloc_free(&acl->bfilter_url);
			stralloc_free(&acl->bfilter_dom);
			stralloc_free(&acl->bfilter_md5);
			stralloc_free(&acl->bfilter_sha1);
			
			stralloc_free(&acl->ufilter);
			stralloc_free(&acl->urfilter);
			stralloc_free(&acl->ufilter_url);
			stralloc_free(&acl->ufilter_dom);
			stralloc_free(&acl->ufilter_md5);
			stralloc_free(&acl->ufilter_sha1);
			
			stralloc_free(&acl->wfilter);
			stralloc_free(&acl->wrfilter);
			stralloc_free(&acl->wfilter_url);
			stralloc_free(&acl->wfilter_dom);
			stralloc_free(&acl->wfilter_md5);
			stralloc_free(&acl->wfilter_sha1);
			
			stralloc_free(&acl->pfilter);
			stralloc_free(&acl->cfilter);
			stralloc_free(&acl->crfilter);
			stralloc_free(&acl->bcfilter);
		}
		break;
	case TYPE_REDIRECT:
	case TYPE_STRING:
		for (i=0; i < array_length(root, sizeof(struct squidwall_string)); i++) {
			string = array_get(root, sizeof(struct squidwall_string), i);
			stralloc_free(&string->filename);
			stralloc_free(&string->strings);
		}
		break;
	case TYPE_PHRASE:
		for (i=0; i < array_length(root, sizeof(struct squidwall_phrase)); i++) {
			phrase = array_get(root, sizeof(struct squidwall_phrase), i);
			stralloc_free(&phrase->filename);
			stralloc_free(&phrase->strings);
		}
		break;
	case TYPE_REGEX:
		for (i=0; i < array_length(root, sizeof(struct squidwall_regex)); i++) {
			regex = array_get(root, sizeof(struct squidwall_regex), i);
			stralloc_free(&regex->filename);
			for (j=0; j < array_length(&regex->regex, sizeof(struct named_regex)); j++) {
				nr = array_get(&regex->regex, sizeof(struct named_regex), j);
				regfree(&nr->r);
				stralloc_free(&nr->n);
			}
			array_reset(&regex->regex);
		}
		break;
	}

	if (array_bytes(root)) {
		array_reset(root);
	}

	return;
}

/* inittialize the strallocs of the current acl */
static void acl_init(void)
{
	INIT_LIST_HEAD(&acl->list);

	stralloc_init(&acl->bfilter);
	stralloc_init(&acl->brfilter);
	stralloc_init(&acl->bfilter_url);
	stralloc_init(&acl->bfilter_dom);
	stralloc_init(&acl->bfilter_md5);
	stralloc_init(&acl->bfilter_sha1);

	stralloc_init(&acl->ufilter);
	stralloc_init(&acl->urfilter);
	stralloc_init(&acl->ufilter_url);
	stralloc_init(&acl->ufilter_dom);
	stralloc_init(&acl->ufilter_md5);
	stralloc_init(&acl->ufilter_sha1);

	stralloc_init(&acl->wfilter);
	stralloc_init(&acl->wrfilter);
	stralloc_init(&acl->wfilter_url);
	stralloc_init(&acl->wfilter_dom);
	stralloc_init(&acl->wfilter_md5);
	stralloc_init(&acl->wfilter_sha1);

	stralloc_init(&acl->pfilter);
	stralloc_init(&acl->cfilter);
	stralloc_init(&acl->crfilter);
	stralloc_init(&acl->bcfilter);
}

/* reset the strallocs of the current acl */
static void acl_reset(void)
{
	INIT_LIST_HEAD(&acl->list);

	stralloc_copys(&acl->bfilter, "");
	stralloc_copys(&acl->brfilter, "");
	stralloc_copys(&acl->bfilter_url, "");
	stralloc_copys(&acl->bfilter_dom, "");
	stralloc_copys(&acl->bfilter_md5, "");
	stralloc_copys(&acl->bfilter_sha1, "");

	stralloc_copys(&acl->ufilter, "");
	stralloc_copys(&acl->urfilter, "");
	stralloc_copys(&acl->ufilter_url, "");
	stralloc_copys(&acl->ufilter_dom, "");
	stralloc_copys(&acl->ufilter_md5, "");
	stralloc_copys(&acl->ufilter_sha1, "");

	stralloc_copys(&acl->wfilter, "");
	stralloc_copys(&acl->wrfilter, "");
	stralloc_copys(&acl->wfilter_url, "");
	stralloc_copys(&acl->wfilter_dom, "");
	stralloc_copys(&acl->wfilter_md5, "");
	stralloc_copys(&acl->wfilter_sha1, "");

	stralloc_copys(&acl->pfilter, "");
	stralloc_copys(&acl->cfilter, "");
	stralloc_copys(&acl->crfilter, "");
	stralloc_copys(&acl->bcfilter, "");
}

/**
 * read_dir2array - read a directory
 * 1. check if '!' was modified
 * 2. if modified, scan the whole directory to an array of type X
 * 3. if the acl is a filtering one, scan the filteroptions
 *
 * @param path   the path, which should be read
 * @param root   pointer of the array
 * @param mtime  the last modification time
 * @param type   the squidwall type: acl, string, regex
 * @return       0 = unchanged, >0 = changed
 */
static int read_dir2array(char *info, char *path, array *root,
		time_t *mtime, struct list_head *list, int type)
{
	struct dirent *d;
	struct stat s;
	stralloc sa;
	DIR *D;
	int i;

	stralloc_init(&sa);
	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/!")) die_nomem();
	if (!stralloc_0(&sa)) die_nomem();

	/* if stat fails, we return ok -> no acl / filter directory */
	if (stat(sa.s, &s) == -1) {
		if (type == TYPE_GLOBAL) {
			/* we make our own global defaults */
			read_acls_globals_default();
		}
		stralloc_free(&sa);
		return 0;
	}
	stralloc_free(&sa);

	if (*mtime == s.st_mtime) {
		/* was not modified... */
		return 0;
	}

	/* was modified... */
	*mtime = s.st_mtime;

	/* free the old array */
	free_array(root,type);

	/* open directory or return zero, which means, no changes to the filters! */
	D=opendir(path); if (!D) return 0;
	i = 0;

	/* read it */
	while ((d=readdir(D))) {

		/* skip '.' and '..' */
		if (d->d_name[0]=='.' && d->d_name[1]==0) continue;
		if (d->d_name[0]=='.' && d->d_name[1]=='.' && d->d_name[2]==0)
			continue;

		/* skip '@' (contains regex options) */
		if (d->d_name[0]=='@' && d->d_name[1]==0) continue;

		/* skip '!' */
		if (d->d_name[0]=='!' && d->d_name[1]==0) continue;

		/* skip also these special files */
		if (str_start(d->d_name, "README")) continue;

		debug(3,3, "reading file %s of %s!", d->d_name, path);

		switch (type) {
		case TYPE_ACL:

			/* filename begins with a false character */
			if (d->d_name[0]!='+' && d->d_name[0]!='-' &&
				d->d_name[0]!='?') continue;

			acl=array_allocate(root, sizeof(struct squidwall_acl), i);
			if (!acl) die_nomem();
			acl_init();
			acl_reset();
			if(!stralloc_copys(&acl->filename, d->d_name+1)) die_nomem();
			acl->filename.s[acl->filename.len]=0;

			/* get names */
			switch (d->d_name[0]) {
			case '+':
				/* ALLOW all */
				acl->flags=ACL_ALLOW;
				break;
			case '-':
				/* DENY all */
				acl->flags=ACL_DENY;
				break;
			case '?':
				/* filtered / scanned */
				acl->flags=ACL_FILTER;
				break;
			}
			break;

		case TYPE_GLOBAL:

			/* filename begins with a false character */
			if (d->d_name[0]!='+' && d->d_name[0]!='-'
				&& d->d_name[0]!='?' && d->d_name[1]!=0) continue;

			acl=array_allocate(root, sizeof(struct squidwall_acl), i);
			if (!acl) die_nomem();
			acl_init();
			acl_reset();

			/* get names */
			switch (d->d_name[0]) {
			case '+':
				/* ALLOW all */
				acl->flags=ACL_ALLOW;
				break;
			case '-':
				/* DENY all */
				acl->flags=ACL_DENY;
				break;
			case '?':
				/* filtered / scanned */
				acl->flags=ACL_FILTER;
				break;
			}
			break;

		case TYPE_REGEX:
			read_acls_regex(root, path, d->d_name);
			break;

		case TYPE_STRING:
			read_acls_string(root, path, d->d_name);
			break;

		case TYPE_PHRASE:
			read_acls_phrase(root, path, d->d_name);
			break;
		
		case TYPE_REDIRECT:
			read_acls_redirects(root, path, d->d_name);
			break;
		}
		i++;
	}
	closedir(D);

	/* initialize the lists */
	switch (type) {
	case TYPE_ACL:
		INIT_LIST_HEAD(&hosts);
		for (i=0; i < array_length(root, sizeof(struct squidwall_acl)); i++) {
			acl = array_get(root, sizeof(struct squidwall_acl), i);
			list_add(&acl->list, list); /* add the acl to the users/hosts list */
			if (acl->flags == ACL_FILTER)
				read_filters_acl(acl, path, acl->filename.s);
		}
		if (i) debug(3,3, "%10s: %d", info, i);
		break;
	case TYPE_GLOBAL:
		//read_acls_checkglobals();
		break;
	case TYPE_REGEX:
		INIT_LIST_HEAD(list);
		for (i=0; i < array_length(root, sizeof(struct squidwall_regex)); i++) {
			regex = array_get(root, sizeof(struct squidwall_regex), i);
			list_add(&regex->list, list);
		}
		if (i) debug(3,3, "%10s: %d", info, i);
		break;
	case TYPE_REDIRECT:
	case TYPE_STRING:
		INIT_LIST_HEAD(list);
		for (i=0; i < array_length(root, sizeof(struct squidwall_string)); i++) {
			string = array_get(root, sizeof(struct squidwall_string), i);
			list_add(&string->list, list);
		}
		if (i) debug(3,3, "%10s: %d", info, i);
		break;
	case TYPE_PHRASE:
		INIT_LIST_HEAD(list);
		for (i=0; i < array_length(root, sizeof(struct squidwall_phrase)); i++) {
			phrase = array_get(root, sizeof(struct squidwall_phrase), i);
			list_add(&phrase->list, list);
		}
		if (i) debug(3,3, "%10s: %d", info, i);
		break;
	}

	return i;
}


/**
 * read_acls - read the acl directories
 */
void read_acls(request_t *r)
{
	int reload=0;

	debug(3,3, "read_acls() started.");

	/* url filter */
	reload += read_dir2array("bfilter", bfilter_dir, &a_bfilter,
			&t_bfilter, &bfilter, TYPE_STRING);
	reload += read_dir2array("brfilter", brfilter_dir, &a_brfilter,
			&t_brfilter, &brfilter, TYPE_REGEX);
	reload += read_dir2array("bfilter_url", bfilter_url_dir, &a_bfilter_url,
			&t_bfilter_url, &bfilter_url, TYPE_STRING);
	reload += read_dir2array("bfilter_dom", bfilter_dom_dir, &a_bfilter_dom,
			&t_bfilter_dom, &bfilter_dom, TYPE_STRING);
	reload += read_dir2array("bfilter_md5", bfilter_md5_dir, &a_bfilter_md5,
			&t_bfilter_md5, &bfilter_md5, TYPE_STRING);
	reload += read_dir2array("bfilter_sha1", bfilter_sha1_dir, &a_bfilter_sha1,
			&t_bfilter_sha1, &bfilter_sha1, TYPE_STRING);

	reload += read_dir2array("ufilter", ufilter_dir, &a_ufilter,
			&t_ufilter, &ufilter, TYPE_STRING);
	reload += read_dir2array("urfilter", urfilter_dir, &a_urfilter,
			&t_urfilter, &urfilter, TYPE_REGEX);
	reload += read_dir2array("ufilter_url", ufilter_url_dir, &a_ufilter_url,
			&t_ufilter_url, &ufilter_url, TYPE_STRING);
	reload += read_dir2array("ufilter_dom", ufilter_dom_dir, &a_ufilter_dom,
			&t_ufilter_dom, &ufilter_dom, TYPE_STRING);
	reload += read_dir2array("ufilter_md5", ufilter_md5_dir, &a_ufilter_md5,
			&t_ufilter_md5, &ufilter_md5, TYPE_STRING);
	reload += read_dir2array("ufilter_sha1", ufilter_sha1_dir, &a_ufilter_sha1,
			&t_ufilter_sha1, &ufilter_sha1, TYPE_STRING);

	reload += read_dir2array("wfilter", wfilter_dir, &a_wfilter,
			&t_wfilter, &wfilter, TYPE_STRING);
	reload += read_dir2array("wrfilter", wrfilter_dir, &a_wrfilter,
			&t_wrfilter, &wrfilter, TYPE_STRING);
	reload += read_dir2array("wfilter_url", wfilter_url_dir, &a_wfilter_url,
			&t_wfilter_url, &wfilter_url, TYPE_STRING);
	reload += read_dir2array("wfilter_dom", wfilter_dom_dir, &a_wfilter_dom,
			&t_wfilter_dom, &wfilter_dom, TYPE_STRING);
	reload += read_dir2array("wfilter_md5", wfilter_md5_dir, &a_wfilter_md5,
			&t_wfilter_md5, &wfilter_md5, TYPE_STRING);
	reload += read_dir2array("wfilter_sha1", wfilter_sha1_dir, &a_wfilter_sha1,
			&t_wfilter_sha1, &wfilter_sha1, TYPE_STRING);

	/* content filter */
	reload += read_dir2array("crfilter", crfilter_dir, &a_crfilter,
			&t_crfilter, &crfilter, TYPE_REGEX);
	reload += read_dir2array("bcfilter", bcfilter_dir, &a_bcfilter,
			&t_bcfilter, &bcfilter, TYPE_REGEX);
	reload += read_dir2array("cfilter", cfilter_dir, &a_cfilter,
			&t_cfilter, &cfilter, TYPE_STRING);
	reload += read_dir2array("pfilter", pfilter_dir, &a_pfilter,
			&t_pfilter, &pfilter, TYPE_PHRASE);

	/* redirects */
	reload += read_dir2array("redirects", redir_dir, &a_redirects,
			&t_redirects, &redirects, TYPE_REDIRECT);

	/* read the acl's */
	reload += read_dir2array("users", user_dir, &a_users,
			&t_users, &users, TYPE_ACL);
	reload += read_dir2array("hosts", host_dir, &a_hosts,
			&t_hosts, &hosts, TYPE_ACL);
	reload += read_dir2array("globals", globals_dir, &a_globals,
			&t_globals, &globals, TYPE_ACL);


	// read_acls_regex(&a_cfilter_todo, cfilter_dir, "!");
	if (reload) {
		memcache_reinit(r);
		debug(3,3, vline);
	}

	debug(3,3, "read_acls() ende.");

	return;
}
